@@ -6,15 +6,35 @@
  */
 
 #include "engine.h"
+#include <algorithm>
+#include <cstdlib>
+#include <fcitx-config/iniparser.h>
+#include <fcitx-config/rawconfig.h>
+#include <fcitx-utils/capabilityflags.h>
+#include <fcitx-utils/charutils.h>
+#include <fcitx-utils/key.h>
+#include <fcitx-utils/keysym.h>
+#include <fcitx-utils/misc.h>
 #include <fcitx-utils/standardpath.h>
+#include <fcitx-utils/textformatflags.h>
 #include <fcitx-utils/utf8.h>
 #include <fcitx/action.h>
+#include <fcitx/addoninstance.h>
 #include <fcitx/candidatelist.h>
+#include <fcitx/event.h>
 #include <fcitx/inputcontext.h>
 #include <fcitx/inputcontextmanager.h>
+#include <fcitx/inputmethodentry.h>
 #include <fcitx/inputpanel.h>
 #include <fcitx/statusarea.h>
+#include <fcitx/text.h>
+#include <fcitx/userinterface.h>
 #include <fcitx/userinterfacemanager.h>
+#include <hangul.h>
+#include <memory>
+#include <stdexcept>
+#include <string>
+#include <utility>
 
 static const char *keyboardId[] = {"2",  "2y", "39", "3f", "3s",
                                    "3y", "32", "ro", "ahn"};
@@ -23,6 +43,8 @@ constexpr auto MAX_LENGTH = 40;
 
 namespace fcitx {
 
+namespace {
+
 const KeyList &selectionKeys() {
     static const KeyList selectionKeys{
         Key(FcitxKey_1), Key(FcitxKey_2), Key(FcitxKey_3), Key(FcitxKey_4),
@@ -31,7 +53,7 @@ const KeyList &selectionKeys() {
     return selectionKeys;
 }
 
-std::string ustringToUTF8(const UString &ustr) {
+std::string ustringToUTF8(const std::u32string &ustr) {
     std::string result;
     for (auto c : ustr) {
         result += utf8::UCS4ToUTF8(c);
@@ -39,7 +61,16 @@ std::string ustringToUTF8(const UString &ustr) {
     return result;
 }
 
-static std::string subUTF8String(const std::string &str, int p1, int p2) {
+std::u32string ucsToUString(const ucschar *str) {
+    std::u32string result;
+    while (*str) {
+        result.push_back(*str);
+        str++;
+    }
+    return result;
+}
+
+std::string subUTF8String(const std::string &str, int p1, int p2) {
     int limit;
     int pos;
     int n;
@@ -66,6 +97,19 @@ static std::string subUTF8String(const std::string &str, int p1, int p2) {
     return std::string(begin, end);
 }
 
+HanjaTable *loadTable() {
+    const auto &sp = fcitx::StandardPath::global();
+    std::string hanjaTxt =
+        sp.locate(fcitx::StandardPath::Type::Data, "libhangul/hanja/hanja.txt");
+    HanjaTable *table = nullptr;
+    if (!hanjaTxt.empty()) {
+        table = hanja_table_load(hanjaTxt.c_str());
+    }
+    return table ? table : hanja_table_load(nullptr);
+}
+
+} // namespace
+
 class HangulCandidate : public CandidateWord {
 public:
     HangulCandidate(HangulEngine *engine, int idx, std::string text)
@@ -95,9 +139,9 @@ public:
             reinterpret_cast<void *>(&HangulState::onTransitionCallback), this);
     }
 
-    static bool onTransitionCallback(HangulInputContext *, ucschar c,
-                                     const ucschar *, void *data) {
-        auto that = static_cast<HangulState *>(data);
+    static bool onTransitionCallback(HangulInputContext * /*unused*/, ucschar c,
+                                     const ucschar * /*unused*/, void *data) {
+        auto *that = static_cast<HangulState *>(data);
         return that->onTransition(c);
     }
 
@@ -105,13 +149,15 @@ public:
         if (!*engine_->config().autoReorder) {
             if (hangul_is_choseong(c)) {
                 if (hangul_ic_has_jungseong(context_.get()) ||
-                    hangul_ic_has_jongseong(context_.get()))
+                    hangul_ic_has_jongseong(context_.get())) {
                     return false;
+                }
             }
 
             if (hangul_is_jungseong(c)) {
-                if (hangul_ic_has_jongseong(context_.get()))
+                if (hangul_ic_has_jongseong(context_.get())) {
                     return false;
+                }
             }
         }
 
@@ -120,18 +166,18 @@ public:
 
     void updateLookupTable(bool checkSurrounding) {
         std::string hanjaKey;
-        LookupMethod lookupMethod = LOOKUP_METHOD_PREFIX;
+        LookupMethod lookupMethod = LookupMethod::LOOKUP_METHOD_PREFIX;
 
         hanjaList_.reset();
 
-        auto hic_preedit = hangul_ic_get_preedit_string(context_.get());
-        UString preedit = preedit_;
-        preedit.append(UString(hic_preedit));
+        const auto *hic_preedit = hangul_ic_get_preedit_string(context_.get());
+        std::u32string preedit = preedit_;
+        preedit.append(ucsToUString(hic_preedit));
         if (!preedit.empty()) {
             auto utf8 = ustringToUTF8(preedit);
             if (*engine_->config().wordCommit || *engine_->config().hanjaMode) {
-                hanjaKey = utf8;
-                lookupMethod = LOOKUP_METHOD_PREFIX;
+                hanjaKey = std::move(utf8);
+                lookupMethod = LookupMethod::LOOKUP_METHOD_PREFIX;
             } else {
                 auto cursorPos = ic_->surroundingText().cursor();
                 auto substr =
@@ -141,9 +187,9 @@ public:
                 if (!substr.empty()) {
                     hanjaKey = substr + utf8;
                 } else {
-                    hanjaKey = utf8;
+                    hanjaKey = std::move(utf8);
                 }
-                lookupMethod = LOOKUP_METHOD_SUFFIX;
+                lookupMethod = LookupMethod::LOOKUP_METHOD_SUFFIX;
             }
         } else if (checkSurrounding) {
 
@@ -151,18 +197,18 @@ public:
                 !ic_->surroundingText().isValid()) {
                 return;
             }
-            auto surroundingStr = ic_->surroundingText().text();
+            const auto &surroundingStr = ic_->surroundingText().text();
             auto cursorPos = ic_->surroundingText().cursor();
             auto anchorPos = ic_->surroundingText().anchor();
             if (cursorPos != anchorPos) {
                 // If we have selection in surrounding text, we use that.
                 hanjaKey = subUTF8String(surroundingStr, cursorPos, anchorPos);
-                lookupMethod = LOOKUP_METHOD_EXACT;
+                lookupMethod = LookupMethod::LOOKUP_METHOD_EXACT;
             } else {
                 hanjaKey =
                     subUTF8String(surroundingStr,
                                   static_cast<int>(cursorPos) - 64, cursorPos);
-                lookupMethod = LOOKUP_METHOD_SUFFIX;
+                lookupMethod = LookupMethod::LOOKUP_METHOD_SUFFIX;
             }
         }
 
@@ -172,22 +218,23 @@ public:
         }
     }
 
-    HanjaList *lookupTable(const std::string key, int method) {
+    HanjaList *lookupTable(const std::string &key, LookupMethod method) {
         HanjaList *list = nullptr;
 
-        if (key.empty())
+        if (key.empty()) {
             return nullptr;
+        }
 
         decltype(&hanja_table_match_exact) func = nullptr;
 
         switch (method) {
-        case LOOKUP_METHOD_EXACT:
+        case LookupMethod::LOOKUP_METHOD_EXACT:
             func = &hanja_table_match_exact;
             break;
-        case LOOKUP_METHOD_PREFIX:
+        case LookupMethod::LOOKUP_METHOD_PREFIX:
             func = &hanja_table_match_prefix;
             break;
-        case LOOKUP_METHOD_SUFFIX:
+        case LookupMethod::LOOKUP_METHOD_SUFFIX:
             func = &hanja_table_match_suffix;
             break;
         }
@@ -195,7 +242,7 @@ public:
             return nullptr;
         }
 
-        if (auto symbolTable = engine_->symbolTable()) {
+        if (auto *symbolTable = engine_->symbolTable()) {
             list = func(symbolTable, key.data());
         }
 
@@ -219,7 +266,8 @@ public:
                 cleanup();
             }
             updateUI();
-            return keyEvent.filterAndAccept();
+            keyEvent.filterAndAccept();
+            return;
         }
 
         auto sym = keyEvent.key().sym();
@@ -229,7 +277,7 @@ public:
         }
 
         KeyStates s;
-        for (auto keyList :
+        for (const auto *keyList :
              {&*engine_->config().hanjaModeToggleKey,
               &*engine_->config().prevPageKey, &*engine_->config().nextPageKey,
               &*engine_->config().prevCandidateKey,
@@ -260,28 +308,33 @@ public:
 
         // Handle candidate selection.
         auto candList = ic_->inputPanel().candidateList();
-        if (candList && candList->size() > 0) {
+        if (candList && !candList->empty()) {
             if (keyEvent.key().checkKeyList(*engine_->config().prevPageKey)) {
                 candList->toPageable()->prev();
                 ic_->updateUserInterface(UserInterfaceComponent::InputPanel);
-                return keyEvent.filterAndAccept();
-            } else if (keyEvent.key().checkKeyList(
-                           *engine_->config().nextPageKey)) {
+                keyEvent.filterAndAccept();
+                return;
+            }
+            if (keyEvent.key().checkKeyList(*engine_->config().nextPageKey)) {
                 candList->toPageable()->next();
                 ic_->updateUserInterface(UserInterfaceComponent::InputPanel);
-                return keyEvent.filterAndAccept();
+                keyEvent.filterAndAccept();
+                return;
             }
 
             if (keyEvent.key().checkKeyList(
                     *engine_->config().prevCandidateKey)) {
                 candList->toCursorMovable()->prevCandidate();
                 ic_->updateUserInterface(UserInterfaceComponent::InputPanel);
-                return keyEvent.filterAndAccept();
-            } else if (keyEvent.key().checkKeyList(
-                           *engine_->config().nextCandidateKey)) {
+                keyEvent.filterAndAccept();
+                return;
+            }
+            if (keyEvent.key().checkKeyList(
+                    *engine_->config().nextCandidateKey)) {
                 candList->toCursorMovable()->nextCandidate();
                 ic_->updateUserInterface(UserInterfaceComponent::InputPanel);
-                return keyEvent.filterAndAccept();
+                keyEvent.filterAndAccept();
+                return;
             }
 
             auto idx = keyEvent.key().keyListIndex(selectionKeys());
@@ -289,18 +342,18 @@ public:
                 if (idx < candList->size()) {
                     candList->candidate(idx).select(ic_);
                 }
-                return keyEvent.filterAndAccept();
+                keyEvent.filterAndAccept();
+                return;
             }
 
             if (keyEvent.key().check(FcitxKey_Return)) {
                 auto idx = candList->cursorIndex();
-                if (idx < 0) {
-                    idx = 0;
-                }
+                idx = std::max(idx, 0);
 
                 if (idx < candList->size()) {
                     candList->candidate(idx).select(ic_);
-                    return keyEvent.filterAndAccept();
+                    keyEvent.filterAndAccept();
+                    return;
                 }
             }
 
@@ -332,7 +385,18 @@ public:
                 flush();
             }
 
-            keyUsed = hangul_ic_process(context_.get(), keyEvent.key().sym());
+            // revert capslock
+            if (keyEvent.rawKey().states().test(KeyState::CapsLock)) {
+                if (sym >= 'A' && sym <= 'z') {
+                    if (charutils::isupper(sym)) {
+                        sym = static_cast<KeySym>(charutils::tolower(sym));
+                    } else {
+                        sym = static_cast<KeySym>(charutils::toupper(sym));
+                    }
+                }
+            }
+
+            keyUsed = hangul_ic_process(context_.get(), sym);
             bool notFlush = false;
 
             const ucschar *str = hangul_ic_get_commit_string(context_.get());
@@ -340,10 +404,8 @@ public:
                 const ucschar *hic_preedit;
 
                 hic_preedit = hangul_ic_get_preedit_string(context_.get());
-                if (hic_preedit != NULL && hic_preedit[0] != 0) {
-                    preedit_.append(UString(str));
-                } else {
-                    preedit_.append(UString(str));
+                preedit_.append(ucsToUString(str));
+                if (hic_preedit == nullptr || hic_preedit[0] == 0) {
                     if (!preedit_.empty()) {
                         auto commit = ustringToUTF8(preedit_);
                         if (!commit.empty()) {
@@ -353,8 +415,8 @@ public:
                     preedit_.clear();
                 }
             } else {
-                if (str != NULL && str[0] != 0) {
-                    auto commit = ustringToUTF8(str);
+                if (str != nullptr && str[0] != 0) {
+                    auto commit = ustringToUTF8(ucsToUString(str));
                     if (!commit.empty()) {
                         ic_->commitString(commit);
                     }
@@ -390,12 +452,13 @@ public:
     void flush() {
         cleanup();
 
-        auto str = hangul_ic_flush(context_.get());
+        const auto *str = hangul_ic_flush(context_.get());
 
-        preedit_ += str;
+        preedit_ += ucsToUString(str);
 
-        if (preedit_.empty())
+        if (preedit_.empty()) {
             return;
+        }
 
         auto utf8 = ustringToUTF8(preedit_);
         if (!utf8.empty()) {
@@ -411,8 +474,11 @@ public:
 
         ic_->inputPanel().reset();
 
-        auto pre1 = ustringToUTF8(preedit_);
-        auto pre2 = hic_preedit ? ustringToUTF8(hic_preedit) : "";
+        std::string pre1 = ustringToUTF8(preedit_);
+        std::string pre2;
+        if (hic_preedit) {
+            pre2 = ustringToUTF8(ucsToUString(hic_preedit));
+        }
 
         if (!pre1.empty() || !pre2.empty()) {
             Text text;
@@ -469,15 +535,16 @@ public:
         hic_preedit = hangul_ic_get_preedit_string(context_.get());
 
         if (!key || !value || !hic_preedit) {
-            return reset();
+            reset();
+            return;
         }
 
         key_len = fcitx::utf8::length(std::string(key));
         preedit_len = preedit_.size();
-        hic_preedit_len = UString(hic_preedit).size();
+        hic_preedit_len = ucsToUString(hic_preedit).size();
 
         bool surrounding = false;
-        if (lastLookupMethod_ == LOOKUP_METHOD_PREFIX) {
+        if (lastLookupMethod_ == LookupMethod::LOOKUP_METHOD_PREFIX) {
             if (preedit_len == 0 && hic_preedit_len == 0) {
                 /* remove surrounding_text */
                 if (key_len > 0) {
@@ -515,7 +582,8 @@ public:
             }
 
             /* remove surrounding_text */
-            if (LOOKUP_METHOD_EXACT != lastLookupMethod_ && key_len > 0) {
+            if (LookupMethod::LOOKUP_METHOD_EXACT != lastLookupMethod_ &&
+                key_len > 0) {
                 ic_->deleteSurroundingText(-key_len, key_len);
                 surrounding = true;
             }
@@ -534,21 +602,10 @@ private:
     InputContext *ic_;
     UniqueCPtr<HangulInputContext, &hangul_ic_delete> context_;
     UniqueCPtr<HanjaList, &hanja_list_delete> hanjaList_;
-    UString preedit_;
+    std::u32string preedit_;
     LookupMethod lastLookupMethod_;
 };
 
-HanjaTable *loadTable() {
-    const auto &sp = fcitx::StandardPath::global();
-    std::string hanjaTxt =
-        sp.locate(fcitx::StandardPath::Type::Data, "libhangul/hanja/hanja.txt");
-    HanjaTable *table = nullptr;
-    if (!hanjaTxt.empty()) {
-        table = hanja_table_load(hanjaTxt.c_str());
-    }
-    return table ? table : hanja_table_load(nullptr);
-}
-
 HangulEngine::HangulEngine(Instance *instance)
     : instance_(instance),
       factory_([this](InputContext &ic) { return new HangulState(this, &ic); }),
@@ -573,7 +630,7 @@ HangulEngine::HangulEngine(Instance *instance)
     instance_->inputContextManager().registerProperty("hangulState", &factory_);
 }
 
-void HangulEngine::activate(const InputMethodEntry &,
+void HangulEngine::activate(const InputMethodEntry & /*entry*/,
                             InputContextEvent &event) {
     event.inputContext()->statusArea().addAction(StatusGroup::InputMethod,
                                                  &action_);
@@ -583,22 +640,24 @@ void HangulEngine::activate(const InputMethodEntry &,
 void HangulEngine::deactivate(const InputMethodEntry &entry,
                               InputContextEvent &event) {
     if (event.type() == EventType::InputContextSwitchInputMethod) {
-        auto state = event.inputContext()->propertyFor(&factory_);
+        auto *state = event.inputContext()->propertyFor(&factory_);
         state->flush();
     }
     reset(entry, event);
 }
 
-void HangulEngine::keyEvent(const InputMethodEntry &, KeyEvent &keyEvent) {
+void HangulEngine::keyEvent(const InputMethodEntry & /*entry*/,
+                            KeyEvent &keyEvent) {
     if (keyEvent.isRelease()) {
         return;
     }
-    auto state = keyEvent.inputContext()->propertyFor(&factory_);
+    auto *state = keyEvent.inputContext()->propertyFor(&factory_);
     state->keyEvent(keyEvent);
 }
 
-void HangulEngine::reset(const InputMethodEntry &, InputContextEvent &event) {
-    auto state = event.inputContext()->propertyFor(&factory_);
+void HangulEngine::reset(const InputMethodEntry & /*entry*/,
+                         InputContextEvent &event) {
+    auto *state = event.inputContext()->propertyFor(&factory_);
     state->reset();
 }
 
@@ -618,9 +677,9 @@ HangulState *HangulEngine::state(InputContext *ic) {
 }
 
 void HangulCandidate::select(InputContext *inputContext) const {
-    auto state = engine_->state(inputContext);
+    auto *state = engine_->state(inputContext);
     state->select(idx_);
 }
 } // namespace fcitx
 
-FCITX_ADDON_FACTORY(fcitx::HangulEngineFactory);
+FCITX_ADDON_FACTORY_V2(hangul, fcitx::HangulEngineFactory);
